local BASE_DISTANCE = 450
local mapscale = FRACUNIT
local MAX_DISTANCE = BASE_DISTANCE * mapscale
local MAX_DISTANCE_THRESHOLD = BASE_DISTANCE * mapscale

-- Sync our gargoyle.
addHook("NetVars", function(sync)
	BASE_DISTANCE = sync(BASE_DISTANCE)
	mapscale = sync(mapscale)
	MAX_DISTANCE = sync(MAX_DISTANCE)
	
	--hcombi = sync(hcombi)
end)

addHook("ThinkFrame", function()
	if leveltime == 2 then
		for player in players.iterate do
			player.gargoyle = nil
		end
	end
end)

-- Return either a gargoyle or a partner's mobj.
hcombi.ReturnPartner = function(player, has_partner)
	if has_partner then return players[player.combi].mo end
	if player.gargoyle == nil then 
		-- summon gargoyle >:)))))
		player.gargoyle = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z + 192*mapscale, MT_GARGOYLE)
	else
		if not player.gargoyle.valid then
			player.gargoyle = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z + 192*mapscale, MT_GARGOYLE)
		end
	end
	return player.gargoyle
end

-- Thinker for summoning the ringbits.
hcombi.RingBitThink = function(player, has_partner)
	local partner = hcombi.ReturnPartner(player, has_partner)
	
	hcombi.PlaceRingBit(player)
end

hcombi.PullVarThink = function(player, has_partner)
	local partner = hcombi.ReturnPartner(player, has_partner)
	local plymo = player.mo
	
	-- set our pull
	plymo.pull_xy = hcombi.cv_combipull.value
	partner.pull_xy = hcombi.cv_combipull.value
	plymo.pull_z = hcombi.cv_combipull.value
	partner.pull_z = hcombi.cv_combipull.value
	
	if player.cmd.buttons & BT_ACCELERATE then
			plymo.pull_xy = $+hcombi.cv_combipull.value/2
		end
	if player.cmd.buttons & BT_BRAKE then
		plymo.pull_xy = $+hcombi.cv_combipull.value
	end
	if plymo.z > plymo.floorz then
		if not has_partner then
			partner.pull_z = 0
			partner.pull_xy = 0
		else
			plymo.pull_z = $+hcombi.cv_combipull.value*2
		end
		plymo.cur_airtime = (plymo.cur_airtime or 0) + 1
		if plymo.cur_airtime - (partner.cur_airtime or 0) > 3*TICRATE then
			hcombi.P_TeleportMobj(plymo, partner.x, partner.y, partner.z + 100*mapscale)
			plymo.cur_airtime = 0
			partner.cur_airtime = 0
		end
	else
		plymo.cur_airtime = 0
			
		if partner.z <= partner.floorz and plymo.z + 200*mapscale > partner.z then
			partner.pull_z = 0
		end	
	end

	if player.combi ~= nil then
		if players[player.combi].cmd.buttons & BT_ACCELERATE then
			partner.pull_xy = $+hcombi.cv_combipull.value/2
		end
		if players[player.combi].cmd.buttons & BT_BRAKE then
			partner.pull_xy = $+hcombi.cv_combipull.value
		end
		if partner.z > partner.floorz and partner.pull_z then
			partner.pull_z = $+hcombi.cv_combipull.value*2
		end
				
		if 
		player.kartstuff[k_growshrinktimer] < -2 and players[player.combi].kartstuff[k_growshrinktimer] >= -2 then
			player.kartstuff[k_growshrinktimer] = -2
		elseif player.kartstuff[k_growshrinktimer] > 0 then
			player.kartstuff[k_growshrinktimer] = max($, players[player.combi].kartstuff[k_growshrinktimer])
		elseif (HugeQuest and player.huge >= 0)
			player.huge = max($, players[player.combi].huge)
		end
		player.mo.destscale = max($, players[player.combi].mo.destscale)
		player.mo.scalespeed = max($, players[player.combi].mo.scalespeed)
		for _, prop in ipairs({k_sneakertimer, k_invincibilitytimer, k_driftboost, k_startboost}) do
			player.kartstuff[prop] = max($, players[player.combi].kartstuff[prop])
		end
		player.realtime = min($, players[player.combi].realtime)
		-- HQ Super interop, since without this existing causes some timer bugs
		if HugeQuest then
			if HQ_SuperCheck(players[player.combi]) and not HQ_SuperCheck(player) then
				-- sync both so that we dont get any weird side effects
				player.huge = players[player.combi].huge
				//player.mo.sparkle = max($, players[player.combi].mo.sparkle)
				player.kartstuff[k_invincibilitytimer] = players[player.combi].kartstuff[k_invincibilitytimer]
				K_PlayPowerGloatSound(player.mo)
				S_ChangeMusic("HQSUPR", true, player)
			end
		end

		if players[player.combi].exiting and not player.exiting then
			player.exiting = players[player.combi].exiting
			P_RestoreMusic(player)
		end
	end
end

hcombi.YankRespawnThink = function(player, has_partner)
	local partner = hcombi.ReturnPartner(player, has_partner)
	local plymo = player.mo
	
	local yank_xy = FixedDiv(partner.pull_xy, plymo.pull_xy+partner.pull_xy)/4
	local yank_z = FixedDiv(partner.pull_z, plymo.pull_xy+partner.pull_z)/4
	
	if not has_partner then
		if player.kartstuff[k_respawn] then
			if (not player.combi_respawn) or (player.mo.momz > -3*mapscale and player.mo.z - player.mo.floorz > 20*mapscale) then
				local dist = FixedMul(mapscale, 60)
					
				-- Let players shift sideways while respawning if they hold drift?
				if player.cmd.buttons & BT_DRIFT then
					local speed = player.cmd.driftturn/100
					P_TryMove(player.mo, player.mo.x - sin(player.mo.angle)*speed, player.mo.y + cos(player.mo.angle)*speed, true)
				end
				
				player.combi_respawn = true
					
				hcombi.P_TeleportMobj(partner, player.mo.x, player.mo.y, player.mo.z + 100*mapscale)
				partner.momx, partner.momy, partner.momz = 0, 0, 0
				partner.angle = player.mo.angle
				
				return
			end
		else
			player.combi_respawn = false
		end
	elseif player.kartstuff[k_respawn] > 1 and (player.kartstuff[k_respawn] < players[player.combi].kartstuff[k_respawn]+1 or players[player.combi].kartstuff[k_respawn] == 0) then
		hcombi.P_TeleportMobj(player.mo, players[player.combi].mo.x, players[player.combi].mo.y, players[player.combi].mo.z + 100*mapscale)
		player.mo.momx, player.mo.momy, player.mo.momz = 0, 0, 0
		player.mo.angle = players[player.combi].mo.angle
		--player.kartstuff[k_hyudorotimer] = max($, 2*TICRATE)
				
		if players[player.combi].kartstuff[k_respawn] > 1 then
			players[player.combi].mo.momx, players[player.combi].mo.momy, players[player.combi].mo.momz = 0, 0, 0
		end
		return
	elseif players[player.combi].kartstuff[k_respawn] > 1 then
		return
	end
end

local endtime = (5*TICRATE+TICRATE/2) + TICRATE/4

hcombi.DistanceThink = function(player, has_partner)
	local partner = hcombi.ReturnPartner(player, has_partner)
	local plymo = player.mo
	
	local yank_xy = FixedDiv(partner.pull_xy, plymo.pull_xy+partner.pull_xy)/4
	local yank_z = FixedDiv(partner.pull_z, plymo.pull_xy+partner.pull_z)/4
	
	local distance = hcombi.P_GetDistance(hcombi.P_GetDistance(player.mo.x - partner.x, player.mo.y - partner.y), player.mo.z - partner.z)
		
	partner.respawning = partner.respawning or 0
	if partner.respawning > 0 then
		partner.momx, partner.momy, partner.momz = 0, 0, 0
		hcombi.P_TeleportMobj(partner, player.mo.x, player.mo.y, player.mo.z + 100*mapscale) 
		partner.respawning = $ - 1
	end
	if distance < MAX_DISTANCE then return end
	-- TELEPORT SPOOF
	-- no more gettiing stuck between the teleporters at orr
	-- or monster monarch
	-- or cosmic eclipse
	-- or literally any map that has a teleporter
	if hcombi.cv_teleportspoof.value ~= 0 then
		if (distance > hcombi.cv_teleportspoof.value*mapscale) then 
			if partner.player and partner.player.deadtimer > 1 then return end
			if not has_partner and player.deadtimer > 1 then return end
			partner.momx, partner.momy, partner.momz = 0, 0, 0
			partner.angle = player.mo.angle
			if partner.player ~= nil then
				partner.respawning = TICRATE
				partner.player.kartstuff[k_respawn] = TICRATE
				partner.player.powers[pw_flashing] = 0
				partner.player.powers[pw_nocontrol] = 0
			end
			hcombi.P_TeleportMobj(partner, player.mo.x, player.mo.y, player.mo.z + 100*mapscale) 
			return
		end
	end
	
	if hcombi.cv_tugrespawn.value == 0 then 
		if partner.player and partner.player.deadtimer > 1 then return end
		if not has_partner and player.deadtimer > 1 then return end 
	end

	distance = $-MAX_DISTANCE
		
	local angle = R_PointToAngle2(player.mo.x, player.mo.y, partner.x, partner.y)
	local v_angle = R_PointToAngle2(0, player.mo.z, R_PointToDist2(player.mo.x, player.mo.y, partner.x, partner.y), partner.z)

	P_Thrust(player.mo, angle, FixedMul(FixedMul(cos(v_angle), yank_xy), distance))
	player.mo.momz = $+FixedMul(FixedMul(sin(v_angle), yank_z), distance)
		
	if not has_partner then
		P_Thrust(partner, angle, -FixedMul(FixedMul(cos(v_angle), FRACUNIT-yank_xy), distance))
		partner.momz = $-FixedMul(FixedMul(sin(v_angle), FRACUNIT-yank_z), distance)
	end
	
	-- Make tugs at the start of the race less crazy
	if leveltime < endtime + TICRATE then
		player.mo.momx = $/3
		player.mo.momy = $/3
	end
end

hcombi.MiscellaneousThinker = function()
	for player in players.iterate do
		-- when a player spectates after partners have been assigned, just summon a gargoyle
		if player.combi == nil then 
			player.combi_p = {name = "anti-loneliness gargoyle", valid = false, mo = player.gargoyle}
			player.has_partner = false
			continue 
		end
		player.combi_p = players[player.combi]
		if player.combi_p == nil or player.combi_p.mo == nil then player.combi = nil continue end
	end
end

hcombi.modScroll = 0

hcombi.PartnerScroll = function()
	if leveltime == 2 then
		--print("reset goes here")
		hcombi.combi_on = hcombi.cv_combi.value
		hcombi.combi_initscroll = -2*TICRATE
		hcombi.modScroll = 0
		for player in players.iterate do
			player.fake_combi = player
			player.old_combi = player
			if hcombi.cv_useoldalgorithm.value == 0 then
				player.combi = nil
				player.has_partner = false
			end
		end
	elseif not hcombi.combi_on then
		return
	elseif leveltime < endtime and leveltime > 2 then
		--if leveltime == START_TIME then S_StartSound(nil, sfx_token) end
		hcombi.combi_initscroll = $ - min((endtime - leveltime)*2/5, TICRATE*2)
		if hcombi.combi_initscroll < 0 then
			hcombi.combi_initscroll = $ + TICRATE*6
			--print(hcombi.combi_initscroll)
		else
			return
		end
		
		-- combi roulette emulation
		local allPlayers = {}
		for player in players.iterate do
			if player.mo and player.mo.valid then
				player.old_combi = player.fake_combi
				table.insert(allPlayers, player)
			else
				player.fake_combi = player
			end
		end
		
		-- oh hey
		while #allPlayers >= 2 do
			local one, two
			-- apparently jart doesnt want these separate or itll screem
			one = table.remove(allPlayers, hcombi.random.randomRange(1,#allPlayers))
			two = table.remove(allPlayers, hcombi.random.randomRange(1,#allPlayers))
			-- these work, apparently.
			local rnd1 = #one
			local rnd2 = #two
			
			-- fake combi is for other stuff
			one.fake_combi = two
			two.fake_combi = one
			
			-- old algorithm stuff
			if hcombi.combi_initscroll >= 200 and hcombi.cv_useoldalgorithm.value == 1 then
				one.fake_combi = one.combi_p
				two.fake_combi = two.combi_p
			end
			
			-- using the newer algorithm is better since itll work with someone speccing
			if hcombi.cv_useoldalgorithm.value == 0 then
				one.combi = rnd2
				two.combi = rnd1
				one.has_partner = true
				two.has_partner = true
			end
		end
		
		-- we are odd!!!!!!!!!!!!
		if #allPlayers then
			allPlayers[1].fake_combi = {valid = false, name = "---"}
			if hcombi.cv_useoldalgorithm.value == 0 then
				allPlayers[1].combi = nil
				allPlayers[1].has_partner = false
			end
		end

		
		S_StartSound(nil, sfx_s1ba)
	end
end

local starttime = 6 * TICRATE + (3 * TICRATE / 4)

addHook("ThinkFrame", function()
	-- do not execute ANYTHING here when combi is not on.
	if leveltime == 2 then
		hcombi.combi_on = hcombi.cv_combi.value
	end
	if not hcombi.combi_on then return end
	if leveltime == 2 then
		pcall(do COM_BufInsertText(server, "karteliminatelast off") end)
	end
	hcombi.PartnerScroll()
	hcombi.MiscellaneousThinker()
	/*
	for player in players.iterate do
		player.mo.spritexscale = sin(FixedAngle((leveltime%360)*FRACUNIT))+FRACUNIT
		player.mo.spriteyscale = cos(FixedAngle((leveltime%360)*FRACUNIT))+FRACUNIT
		player.mo.rollangle = FixedAngle((leveltime%360)*FRACUNIT)
		player.mo.rollmodel = true
	end
	*/
	if leveltime < endtime then return end
	hcombi.combi_initscroll = $/2
	if leveltime == endtime then S_StartSound(nil, sfx_token) end
	
	-- ok now we're here, lets talk a bit about fickle's combi mod.
	-- desyncs are prevalent when using this in replays, as it has a lot of problems.
	-- i would normally just fix it but it is non reusable
	
	-- this uses some of the code from fickle's but it is not a direct edit.
	-- this is somewhat of a reimplementation and fixes some of the original's quirks.
	-- otherwise this doesnt differ all too much
	
	-- fickle if youre reading this thank you for the amount of chaos you have brought to this community
	
	-- max thanks to bladesinger from kru for doing a small combi ring marathon
	-- i wouldnt have thought of this without you man
	
	-- - haya
	
	BASE_DISTANCE = hcombi.cv_combidist.value
	mapscale = (mapheaderinfo[gamemap] and mapheaderinfo[gamemap].mobj_scale or FRACUNIT)
	MAX_DISTANCE = BASE_DISTANCE * mapscale
	
	for player in players.iterate do
		if player.spectator then continue end
		--player.mo.rollangle = $1 - ((ANGLE_11hh * 2)/2);
		--player.mo.rollmodel = true;
		--player.mo.spritexscale = $ + (FRACUNIT)
		hcombi.RingBitThink(player, player.has_partner)
		hcombi.PullVarThink(player, player.has_partner)
		hcombi.YankRespawnThink(player, player.has_partner)
		hcombi.DistanceThink(player, player.has_partner)
	end
end)

-- Make rankings display per pair instead of per individual?
addHook("ThinkFrame", do
	if not hcombi.combi_on then return end

	local individual_rank = {}
	--print("---------------")
	for player in players.iterate do
		player.combi_rank = nil
		if not player.spectator then
			--print(player.real_rank)
			player.real_rank = $ or player.kartstuff[k_position]
			
			local spot = 1
			while individual_rank[spot] and individual_rank[spot].real_rank <= player.real_rank do
				spot = $+1
			end

			table.insert(individual_rank, spot, player)
		end
	end
	
	local rank = 1
	local team_rank = 0
	
	while individual_rank[rank] do
		local player = individual_rank[rank]
		player.real_rank = nil
		if player.combi ~= nil and players[player.combi] and players[player.combi].valid and players[player.combi].combi_rank then
			player.combi_rank = players[player.combi].combi_rank
		else
			team_rank = $+1
			player.combi_rank = team_rank
		end
		
		if player.combi_rank ~= player.old_combi_rank then
			player.combi_rank_cooldown = 10
			player.old_combi_rank = player.combi_rank
		elseif player.combi_rank_cooldown > 0 then
			player.combi_rank_cooldown = $-1
		end
		
		-- This handles HUD display of the ranking, I think!
		player.kartstuff[k_position] = player.combi_rank
		player.kartstuff[k_positiondelay] = player.combi_rank_cooldown
		
		rank = $+1
	end
end)

-- Force rankings to be mostly-correct on player thinkers. This causes the distance-checking code for giving SPBs to work properly.
addHook("MobjThinker", do
	if not combi_on then return end
	
	--print("-----")
	for player in players.iterate do
		-- This "real_rank" juggling is necessary to make sure we don't break the team position display.
		if not player.real_rank then
			player.real_rank = player.kartstuff[k_position]
		end

		if player.combi_rank then
			player.kartstuff[k_position] = player.combi_rank
		end
	end
end, MT_PLAYER)

